# coding: utf-8
from __future__ import absolute_import, division, print_function

import pytest

from parver._helpers import force_tuple, kwonly_args


@pytest.mark.parametrize('arg, expected', [
    (1, (1,)),
    ([1], (1,)),
    ((1, 2), (1, 2)),
    ([1, 2], (1, 2)),
    # range is a Sequence
    (range(1, 3), (1, 2)),
    # An iterable that is not also a sequence
    ((x for x in range(1, 3)), (1, 2)),
])
def test_force_tuple(arg, expected):
    assert force_tuple(arg) == expected


def fn1(a, **kwargs):
    kwargs, b, c = kwonly_args(kwargs, ('b',), (('c', 5),))
    return a, b, c


def fn2(a, *args, **kwargs):
    kwargs, b = kwonly_args(kwargs, ('b',), leftovers=True)
    return a, b, kwargs


@pytest.mark.parametrize('fn, args, kwargs, result', [
    (fn1, (1,), dict(b=2), (1, 2, 5)),
    (fn1, (1,), dict(b=2, c=3), (1, 2, 3)),
    (fn1, (), dict(b=2), TypeError),
    (fn1, (1, 2), dict(), TypeError),
    (fn1, (1,), dict(c=2), TypeError),
    (fn1, (1,), dict(b=2, d=4), TypeError),
    (fn2, (1,), dict(b=2), (1, 2, {})),
    (fn2, (1,), dict(b=2, c=3), (1, 2, dict(c=3))),
])
def test_kwonly_args(fn, args, kwargs, result):
    """Based on the snippet by Eric Snow
    http://code.activestate.com/recipes/577940

    SPDX-License-Identifier: MIT
    """
    try:
        isexception = issubclass(result, Exception)
    except TypeError:
        isexception = False

    if isexception:
        with pytest.raises(result):
            fn(*args, **kwargs)
    else:
        assert fn(*args, **kwargs) == result



# coding: utf-8
from __future__ import absolute_import, division, print_function

import os

from hypothesis import settings, Verbosity

settings.register_profile('ci', max_examples=1000)
settings.register_profile('debug', verbosity=Verbosity.verbose)
settings.load_profile(os.getenv('HYPOTHESIS_PROFILE', 'default'))



# coding: utf-8
from __future__ import absolute_import, division, print_function

import itertools
import operator
import re
from collections import Sequence
from functools import partial

import attr
import six
from attr.validators import in_, instance_of, optional

from ._helpers import UNSET, Infinity, doc_signature, force_tuple, kwonly_args
from ._parse import parse
from . import _segments as segment


POST_TAGS = {'post', 'rev', 'r'}
SEPS = {'.', '-', '_'}
PRE_TAGS = {'c', 'rc', 'alpha', 'a', 'beta', 'b', 'preview', 'pre'}


def unset_or(validator):
    def validate(inst, attr, value):
        if value is UNSET:
            return

        validator(inst, attr, value)
    return validate


validate_post_tag = unset_or(optional(in_(POST_TAGS)))
validate_pre_tag = optional(in_(PRE_TAGS))
validate_sep = optional(in_(SEPS))
validate_sep_or_unset = unset_or(optional(in_(SEPS)))
is_bool = instance_of(bool)
is_int = instance_of(six.integer_types)
is_str = instance_of(six.string_types)
is_seq = instance_of(Sequence)


@attr.s(frozen=True, repr=False, cmp=False)
class Version(object):
    """

    :param release: Numbers for the release segment.
    :type release: int or tuple[int]

    :param v: Optional preceding v character.
    :type v: bool

    :param epoch: `Version epoch`_. Implicitly zero but hidden by default.
    :type epoch: int

    :param pre_tag: `Pre-release`_ identifier, typically `a`, `b`, or `rc`.
        Required to signify a pre-release.
    :type pre_tag: str

    :param pre: `Pre-release`_ number. May be `None` to signify an
        `implicit pre-release number`_.
    :type pre: int

    :param post: `Post-release`_ number. May be `None` to signify an
        `implicit post release number`_.
    :type post: int

    :param dev: `Developmental release`_ number. May be `None` to signify an
        `implicit development release number`_.
    :type dev: int

    :param local: `Local version`_ segment.
    :type local:

    :param pre_sep1: Specify an alternate separator before the pre-release
        segment. The normal form is `None`.
    :type pre_sep1: str

    :param pre_sep2: Specify an alternate separator between the identifier and
        number. The normal form is ``'.'``.
    :type pre_sep2: str

    :param post_sep1: Specify an alternate separator before the post release
        segment. The normal form is ``'.'``.
    :type post_sep1: str

    :param post_sep2: Specify an alternate separator between the identifier and
        number. The normal form is ``'.'``.
    :type post_sep2: str

    :param dev_sep: Specify an alternate separator before the development
        release segment. The normal form is ``'.'``.
    :type dev_sep: str

    :param post_tag: Specify alternate post release identifier `rev` or `r`.
        May be `None` to signify an `implicit post release`_.
    :type post_tag: str

    .. note:: The attributes below are not equal to the parameters passed to
        the initialiser!

        The main difference is that implicit numbers become `0` and set the
        corresponding `_implicit` attribute:

        .. doctest::

            >>> v = Version(release=1, post=None)
            >>> str(v)
            '1.post'
            >>> v.post
            0
            >>> v.post_implicit
            True

    .. attribute:: release

        A tuple of integers giving the components of the release segment of
        this :class:`Version` instance; that is, the ``1.2.3`` part of the
        version number, including trailing zeroes but not including the epoch
        or any prerelease/development/postrelease suffixes

    .. attribute:: v

        Whether this :class:`Version` instance includes a preceding v character.

    .. attribute:: epoch

        An integer giving the version epoch of this :class:`Version` instance.
        :attr:`epoch_implicit` may be `True` if this number is zero.

    .. attribute:: pre_tag

        If this :class:`Version` instance represents a pre-release, this
        attribute will be the pre-release identifier. One of `a`, `b`, `rc`,
        `c`, `alpha`, `beta`, `preview`, or `pre`.

        **Note:** you should not use this attribute to check or compare
        pre-release identifiers. Use :meth:`is_alpha`, :meth:`is_beta`, and
        :meth:`is_release_candidate` instead.

    .. attribute:: pre

        If this :class:`Version` instance represents a pre-release, this
        attribute will be the pre-release number. If this instance is not a
        pre-release, the attribute will be `None`. :attr:`pre_implicit` may be
        `True` if this number is zero.

    .. attribute:: post

        If this :class:`Version` instance represents a postrelease, this
        attribute will be the postrelease number (an integer); otherwise, it
        will be `None`. :attr:`post_implicit` may be `True` if this number
        is zero.

    .. attribute:: dev

        If this :class:`Version` instance represents a development release,
        this attribute will be the development release number (an integer);
        otherwise, it will be `None`. :attr:`dev_implicit` may be `True` if this
        number is zero.

    .. attribute:: local

        A string representing the local version portion of this :class:`Version`
        instance if it has one, or ``None`` otherwise.

    .. attribute:: pre_sep1

        The separator before the pre-release identifier.

    .. attribute:: pre_sep2

        The seperator between the pre-release identifier and number.

    .. attribute:: post_sep1

        The separator before the post release identifier.

    .. attribute:: post_sep2

        The seperator between the post release identifier and number.

    .. attribute:: dev_sep

        The separator before the develepment release identifier.

    .. attribute:: post_tag

        If this :class:`Version` instance represents a post release, this
        attribute will be the post release identifier. One of `post`, `rev`,
        `r`, or `None` to represent an implicit post release.

    .. _`Version epoch`: https://www.python.org/dev/peps/pep-0440/#version-epochs
    .. _`Pre-release`: https://www.python.org/dev/peps/pep-0440/#pre-releases
    .. _`implicit pre-release number`: https://www.python.org/dev/peps/
        pep-0440/#implicit-pre-release-number
    .. _`Post-release`: https://www.python.org/dev/peps/pep-0440/#post-releases
    .. _`implicit post release number`: https://www.python.org/dev/peps/
        pep-0440/#implicit-post-release-number
    .. _`Developmental release`: https://www.python.org/dev/peps/pep-0440/
        #developmental-releases
    .. _`implicit development release number`: https://www.python.org/dev/peps/
        pep-0440/#implicit-development-release-number
    .. _`Local version`: https://www.python.org/dev/peps/pep-0440/
        #local-version-identifiers
    .. _`implicit post release`: https://www.python.org/dev/peps/pep-0440/
        #implicit-post-releases

    """
    release = attr.ib(converter=force_tuple, validator=is_seq)
    v = attr.ib(default=False, validator=is_bool)
    epoch = attr.ib(default=None, validator=optional(is_int))
    pre_tag = attr.ib(default=None, validator=validate_pre_tag)
    pre = attr.ib(default=None, validator=optional(is_int))
    post = attr.ib(default=UNSET, validator=unset_or(optional(is_int)))
    dev = attr.ib(default=UNSET, validator=unset_or(optional(is_int)))
    local = attr.ib(default=None, validator=optional(is_str))

    pre_sep1 = attr.ib(default=None, validator=validate_sep)
    pre_sep2 = attr.ib(default=None, validator=validate_sep)
    post_sep1 = attr.ib(default=UNSET, validator=validate_sep_or_unset)
    post_sep2 = attr.ib(default=UNSET, validator=validate_sep_or_unset)
    dev_sep = attr.ib(default=UNSET, validator=validate_sep_or_unset)
    post_tag = attr.ib(default=UNSET, validator=validate_post_tag)

    epoch_implicit = attr.ib(default=False, init=False)
    pre_implicit = attr.ib(default=False, init=False)
    post_implicit = attr.ib(default=False, init=False)
    dev_implicit = attr.ib(default=False, init=False)
    _key = attr.ib(init=False)

    def __attrs_post_init__(self):
        set = partial(object.__setattr__, self)

        if self.epoch is None:
            set('epoch', 0)
            set('epoch_implicit', True)

        if self.pre_tag is not None and self.pre is None:
            set('pre', 0)
            set('pre_implicit', True)

        if self.pre is not None and self.pre_tag is None:
            raise ValueError('Must set pre_tag if pre is given.')

        if (self.pre_tag is None and
                (self.pre_sep1 is not None or self.pre_sep2 is not None)):
            raise ValueError('Cannot set pre_sep1 or pre_sep2 without pre_tag.')

        if self.post_tag is None:
            if self.post is UNSET or self.post is None:
                raise ValueError(
                    "Implicit post releases (post_tag=None) require a numerical "
                    "value for 'post' argument.")

            if self.post_sep1 is not UNSET or self.post_sep2 is not UNSET:
                raise ValueError(
                    'post_sep1 and post_sep2 cannot be set for implicit post '
                    'releases (post_tag=None)')

            if self.pre_implicit:
                raise ValueError(
                    'post_tag cannot be None with an implicit pre-release '
                    '(pre=None).')

            set('post_sep1', '-')

        if self.post is not UNSET:
            if self.post_tag is UNSET:
                set('post_tag', 'post')
            if self.post is None:
                set('post_implicit', True)
                set('post', 0)

        if self.post_tag is not UNSET and self.post is UNSET:
            set('post_implicit', True)
            set('post', 0)

        if self.dev is None:
            set('dev_implicit', True)
            set('dev', 0)

        if self.post is UNSET:
            set('post', None)

        if self.post_tag is UNSET:
            set('post_tag', None)

        if self.post_sep1 is UNSET:
            set('post_sep1', None if self.post is None else '.')
        if self.post_sep2 is UNSET:
            set('post_sep2', None)

        if self.dev is UNSET:
            set('dev', None)

        if self.dev_sep is UNSET:
            set('dev_sep', None if self.dev is None else '.')

        assert self.post_sep1 is not UNSET
        assert self.post_sep2 is not UNSET

        set('_key', _cmpkey(
            self.epoch,
            self.release,
            _normalize_pre_tag(self.pre_tag),
            self.pre,
            self.post,
            self.dev,
            self.local,
        ))

    @classmethod
    def parse(cls, version, strict=False):
        """
        :param version: Version number as defined in `PEP 440`_.
        :type version: str

        :param strict: Enable strict parsing of the canonical PEP 440 format.
        :type strict: bool

        .. _`PEP 440`: https://www.python.org/dev/peps/pep-0440/

        :raises ParseError: If version is not valid for the given value of
            `strict`.

        .. doctest::
            :options: -IGNORE_EXCEPTION_DETAIL

            >>> Version.parse('1.dev')
            <Version '1.dev'>
            >>> Version.parse('1.dev', strict=True)
            Traceback (most recent call last):
              ...
            parver._parse.ParseError: Expected int at position (1, 6) => '1.dev*'.
        """
        segments = parse(version, strict=strict)

        kwargs = dict()

        for s in segments:
            if isinstance(s, segment.Epoch):
                kwargs['epoch'] = s.value
            elif isinstance(s, segment.Release):
                kwargs['release'] = s.value
            elif isinstance(s, segment.Pre):
                kwargs['pre'] = s.value
                kwargs['pre_tag'] = s.tag
                kwargs['pre_sep1'] = s.sep1
                kwargs['pre_sep2'] = s.sep2
            elif isinstance(s, segment.Post):
                kwargs['post'] = s.value
                kwargs['post_tag'] = s.tag
                kwargs['post_sep1'] = s.sep1
                kwargs['post_sep2'] = s.sep2
            elif isinstance(s, segment.Dev):
                kwargs['dev'] = s.value
                kwargs['dev_sep'] = s.sep
            elif isinstance(s, segment.Local):
                kwargs['local'] = s.value
            elif isinstance(s, segment.V):
                kwargs['v'] = True
            else:
                raise TypeError('Unexpected segment: {}'.format(segment))

        return cls(**kwargs)

    def normalize(self):
        return Version(
            release=self.release,
            epoch=None if self.epoch == 0 else self.epoch,
            pre_tag=_normalize_pre_tag(self.pre_tag),
            pre=self.pre,
            post=UNSET if self.post is None else self.post,
            dev=UNSET if self.dev is None else self.dev,
            local=self.local
        )

    def __str__(self):
        parts = []

        if self.v:
            parts.append('v')

        if not self.epoch_implicit:
            parts.append('{}!'.format(self.epoch))

        parts.append('.'.join(str(x) for x in self.release))

        if self.pre_tag is not None:
            if self.pre_sep1:
                parts.append(self.pre_sep1)
            parts.append(self.pre_tag)
            if self.pre_sep2:
                parts.append(self.pre_sep2)
            if not self.pre_implicit:
                parts.append(str(self.pre))

        if self.post_tag is None and self.post is not None:
            parts.append('-{}'.format(self.post))
        elif self.post_tag is not None:
            if self.post_sep1:
                parts.append(self.post_sep1)
            parts.append(self.post_tag)
            if self.post_sep2:
                parts.append(self.post_sep2)
            if not self.post_implicit:
                parts.append(str(self.post))

        if self.dev is not None:
            if self.dev_sep is not None:
                parts.append(self.dev_sep)
            parts.append('dev')
            if not self.dev_implicit:
                parts.append(str(self.dev))

        if self.local is not None:
            parts.append('+{}'.format(self.local))

        return ''.join(parts)

    def __repr__(self):
        return '<{} {!r}>'.format(self.__class__.__name__, str(self))

    def __hash__(self):
        return hash(self._key)

    def __lt__(self, other):
        return self._compare(other, operator.lt)

    def __le__(self, other):
        return self._compare(other, operator.le)

    def __eq__(self, other):
        return self._compare(other, operator.eq)

    def __ge__(self, other):
        return self._compare(other, operator.ge)

    def __gt__(self, other):
        return self._compare(other, operator.gt)

    def __ne__(self, other):
        return self._compare(other, operator.ne)

    def _compare(self, other, method):
        if not isinstance(other, Version):
            return NotImplemented

        return method(self._key, other._key)

    @property
    def public(self):
        """A string representing the public version portion of this
        :class:`Version` instance.
        """
        return str(self).split('+', 1)[0]

    def base_version(self):
        """Return a new :class:`Version` instance for the base version of the
        current instance. The base version is the public version of the project
        without any pre or post release markers.

        See also: :meth:`clear` and :meth:`replace`.
        """
        return self.clear(pre=True, post=True, dev=True).replace(local=None)

    @property
    def is_prerelease(self):
        """A boolean value indicating whether this :class:`Version` instance
        represents a pre-release and/or development release.
        """
        return self.dev is not None or self.pre is not None

    @property
    def is_alpha(self):
        """A boolean value indicating whether this :class:`Version` instance
        represents an alpha pre-release.
        """
        return _normalize_pre_tag(self.pre_tag) == 'a'

    @property
    def is_beta(self):
        """A boolean value indicating whether this :class:`Version` instance
        represents a beta pre-release.
        """
        return _normalize_pre_tag(self.pre_tag) == 'b'

    @property
    def is_release_candidate(self):
        """A boolean value indicating whether this :class:`Version` instance
        represents a release candidate pre-release.
        """
        return _normalize_pre_tag(self.pre_tag) == 'rc'

    @property
    def is_postrelease(self):
        """A boolean value indicating whether this :class:`Version` instance
        represents a post-release.
        """
        return self.post is not None

    @property
    def is_devrelease(self):
        """A boolean value indicating whether this :class:`Version` instance
        represents a development release.
        """
        return self.dev is not None

    def _attrs_as_init(self):
        d = attr.asdict(self, filter=lambda attr, _: attr.init)

        if self.epoch_implicit:
            d['epoch'] = None

        if self.pre_implicit:
            d['pre'] = None

        if self.post_implicit:
            d['post'] = None

        if self.dev_implicit:
            d['dev'] = None

        if self.pre is None:
            del d['pre']
            del d['pre_tag']
            del d['pre_sep1']
            del d['pre_sep2']

        if self.post is None:
            del d['post']
            del d['post_tag']
            del d['post_sep1']
            del d['post_sep2']
        elif self.post_tag is None:
            del d['post_sep1']
            del d['post_sep2']

        if self.dev is None:
            del d['dev']
            del d['dev_sep']

        return d

    def clear(self, pre=False, post=False, dev=False):
        """Like :meth:`replace`, but has the ability to **remove** pre-release,
        post release, and development release segments.

        See also: :meth:`base_version`.
        """
        d = self._attrs_as_init()

        if pre:
            d.pop('pre', None)
            d.pop('pre_tag', None)
            d.pop('pre_sep1', None)
            d.pop('pre_sep2', None)

        if post:
            d.pop('post', None)
            d.pop('post_tag', None)
            d.pop('post_sep1', None)
            d.pop('post_sep2', None)

        if dev:
            d.pop('dev', None)
            d.pop('dev_sep', None)

        return Version(**d)

    def replace(self, **kwargs):
        """Return a new :class:`Version` instance with the same attributes,
        except for those given as keyword arguments. Arguments have the same
        meaning as they do when constructing a new :class:`Version` instance
        manually.

        .. warning::

            Be careful! :class:`Version` treats `None` as an implicit zero, so
            pre-release, post release and development releases cannot be
            cleared using this method:

            .. doctest::

                >>> Version.parse('1.3.post0').replace(post=None)
                <Version '1.3.post'>
                >>> Version.parse('1.3').replace(post=None)
                <Version '1.3.post'>

            Use :meth:`clear` instead:

            .. doctest::

                >>> Version.parse('1.3.post0').clear(post=True)
                <Version '1.3'>

        """
        d = self._attrs_as_init()

        if kwargs.get('post_tag', UNSET) is None:
            # ensure we don't carry over separators for new implicit post
            # release. By popping from d, there will still be an error if the
            # user tries to set them in kwargs
            d.pop('post_sep1', None)
            d.pop('post_sep2', None)

        d.update(kwargs)
        return Version(**d)

    def _set_release(self, index, value=None, bump=True):
        if not isinstance(index, int):
            raise TypeError('index must be an integer')

        if index < 0:
            raise ValueError('index cannot be negative')

        release = list(self.release)
        new_len = index + 1

        if len(release) < new_len:
            release.extend(itertools.repeat(0, new_len - len(release)))

        def new_parts(i, n):
            if i < index:
                return n
            if i == index:
                if value is None:
                    return n + 1
                return value
            if bump:
                return 0
            return n

        release = itertools.starmap(new_parts, enumerate(release))
        return self.replace(release=release)

    @doc_signature('(*, index)')
    def bump_release(self, **kwargs):
        """Return a new :class:`Version` instance with the release number
        bumped at the given `index`.

        :param index: Index of the release number tuple to bump. It is not
            limited to the current size of the tuple. Intermediate indices will
            be set to zero.
        :type index: int

        :raises TypeError: `index` is not an integer.
        :raises ValueError: `index` is negative.

        .. doctest::

            >>> v = Version.parse('1.4')
            >>> v.bump_release(index=0)
            <Version '2.0'>
            >>> v.bump_release(index=1)
            <Version '1.5'>
            >>> v.bump_release(index=2)
            <Version '1.4.1'>
            >>> v.bump_release(index=3)
            <Version '1.4.0.1'>

        .. seealso::

            For more control over the value that is bumped to, see
            :meth:`bump_release_to`.

            For fine-grained control, :meth:`set_release` may be used to set
            the value at a specific index without setting subsequenct indices
            to zero.
        """
        _, index = kwonly_args(kwargs, ('index',))
        return self._set_release(index=index)

    @doc_signature('(*, index, value)')
    def bump_release_to(self, **kwargs):
        """Return a new :class:`Version` instance with the release number
        bumped at the given `index` to `value`. May be used for versioning
        schemes such as `CalVer`_.

        .. _`CalVer`: https://calver.org

        :param index: Index of the release number tuple to bump. It is not
            limited to the current size of the tuple. Intermediate indices will
            be set to zero.
        :type index: int
        :param value: Value to bump to. This may be any value, but subsequent
            indices will be set to zero like a normal version bump.
        :type value: int

        :raises TypeError: `index` is not an integer.
        :raises ValueError: `index` is negative.

        .. testsetup::

            import datetime

        .. doctest::

            >>> v = Version.parse('18.4')
            >>> v.bump_release_to(index=0, value=20)
            <Version '20.0'>
            >>> v.bump_release_to(index=1, value=10)
            <Version '18.10'>

        For a project using `CalVer`_ with format ``YYYY.MM.MICRO``, this
        method could be used to set the date parts:

        .. doctest::

            >>> v = Version.parse('2018.4.1')
            >>> v = v.bump_release_to(index=0, value=2018)
            >>> v = v.bump_release_to(index=1, value=10)
            >>> v
            <Version '2018.10.0'>

        .. seealso::

            For typical use cases, see :meth:`bump_release`.

            For fine-grained control, :meth:`set_release` may be used to set
            the value at a specific index without setting subsequenct indices
            to zero.
        """
        _, index, value = kwonly_args(kwargs, ('index', 'value'))
        return self._set_release(index=index, value=value)

    @doc_signature('(*, index, value)')
    def set_release(self, **kwargs):
        """Return a new :class:`Version` instance with the release number
        at the given `index` set to `value`.

        :param index: Index of the release number tuple to set. It is not
            limited to the current size of the tuple. Intermediate indices will
            be set to zero.
        :type index: int
        :param value: Value to set.
        :type value: int

        :raises TypeError: `index` is not an integer.
        :raises ValueError: `index` is negative.

        .. doctest::

            >>> v = Version.parse('1.2.3')
            >>> v.set_release(index=0, value=3)
            <Version '3.2.3'>
            >>> v.set_release(index=1, value=4)
            <Version '1.4.3'>

        .. seealso::

            For typical use cases, see :meth:`bump_release`.
        """
        _, index, value = kwonly_args(kwargs, ('index', 'value'))
        return self._set_release(index=index, value=value, bump=False)

    def bump_pre(self, tag=None):
        """Return a new :class:`Version` instance with the pre-release number
        bumped.

        :param tag: Pre-release tag. Required if not already set.
        :type tag: str

        :raises ValueError: Trying to call ``bump_pre(tag=None)`` on a
            :class:`Version` instance that is not already a pre-release.
        :raises ValueError: Calling the method with a `tag` not equal to the
            current :attr:`post_tag`. See :meth:`replace` instead.

        .. doctest::

            >>> Version.parse('1.4').bump_pre('a')
            <Version '1.4a0'>
            >>> Version.parse('1.4b1').bump_pre()
            <Version '1.4b2'>
        """
        pre = 0 if self.pre is None else self.pre + 1

        if self.pre_tag is None:
            if tag is None:
                raise ValueError(
                    "Cannot bump without pre_tag. Use .bump_pre('<tag>')")
        else:
            # This is an error because different tags have different meanings
            if tag is not None and self.pre_tag != tag:
                raise ValueError(
                    'Cannot bump with pre_tag mismatch ({0} != {1}). Use '
                    '.replace(pre_tag={1!r})'.format(self.pre_tag, tag))
            tag = self.pre_tag

        return self.replace(pre=pre, pre_tag=tag)

    def bump_post(self, tag=UNSET):
        """Return a new :class:`Version` instance with the post release number
        bumped.

        :param tag: Post release tag. Will preserve the current tag by default,
            or use `post` if the instance is not already a post release.
        :type tag: str

        .. doctest::

            >>> Version.parse('1.4').bump_post()
            <Version '1.4.post0'>
            >>> Version.parse('1.4.post0').bump_post(tag=None)
            <Version '1.4-1'>
            >>> Version.parse('1.4_post-1').bump_post(tag='rev')
            <Version '1.4_rev-2'>
        """
        post = 0 if self.post is None else self.post + 1
        if tag is UNSET and self.post is not None:
            tag = self.post_tag
        return self.replace(post=post, post_tag=tag)

    def bump_dev(self):
        """Return a new :class:`Version` instance with the development release
        number bumped.

        .. doctest::

            >>> Version.parse('1.4').bump_dev()
            <Version '1.4.dev0'>
            >>> Version.parse('1.4_dev1').bump_dev()
            <Version '1.4_dev2'>
        """
        dev = 0 if self.dev is None else self.dev + 1
        return self.replace(dev=dev)


def _normalize_pre_tag(pre_tag):
    if pre_tag is None:
        return None

    if pre_tag == 'alpha':
        pre_tag = 'a'
    elif pre_tag == 'beta':
        pre_tag = 'b'
    elif pre_tag in {'c', 'pre', 'preview'}:
        pre_tag = 'rc'

    return pre_tag


def _cmpkey(epoch, release, pre_tag, pre_num, post, dev, local):
    # When we compare a release version, we want to compare it with all of the
    # trailing zeros removed. So we'll use a reverse the list, drop all the now
    # leading zeros until we come to something non zero, then take the rest
    # re-reverse it back into the correct order and make it a tuple and use
    # that for our sorting key.
    release = tuple(
        reversed(list(
            itertools.dropwhile(
                lambda x: x == 0,
                reversed(release),
            )
        ))
    )

    pre = pre_tag, pre_num

    # We need to "trick" the sorting algorithm to put 1.0.dev0 before 1.0a0.
    # We'll do this by abusing the pre segment, but we _only_ want to do this
    # if there is not a pre or a post segment. If we have one of those then
    # the normal sorting rules will handle this case correctly.
    if pre_num is None and post is None and dev is not None:
        pre = -Infinity
    # Versions without a pre-release (except as noted above) should sort after
    # those with one.
    elif pre_num is None:
        pre = Infinity

    # Versions without a post segment should sort before those with one.
    if post is None:
        post = -Infinity

    # Versions without a development segment should sort after those with one.
    if dev is None:
        dev = Infinity

    if local is None:
        # Versions without a local segment should sort before those with one.
        local = -Infinity
    else:
        # Versions with a local segment need that segment parsed to implement
        # the sorting rules in PEP440.
        # - Alpha numeric segments sort before numeric segments
        # - Alpha numeric segments sort lexicographically
        # - Numeric segments sort numerically
        # - Shorter versions sort before longer versions when the prefixes
        #   match exactly
        local = tuple(
            (i, "") if isinstance(i, int) else (-Infinity, i)
            for i in _parse_local_version(local)
        )

    return epoch, release, pre, post, dev, local


_local_version_separators = re.compile(r"[._-]")


def _parse_local_version(local):
    """
    Takes a string like abc.1.twelve and turns it into ("abc", 1, "twelve").
    """
    if local is not None:
        return tuple(
            part.lower() if not part.isdigit() else int(part)
            for part in _local_version_separators.split(local)
        )



# coding: utf-8
from __future__ import absolute_import, division, print_function


def autodoc_process_signature(
        app, what, name, obj, options, signature, return_annotation,
):
    try:
        signature = obj.__parverdoc_signature__
    except AttributeError:
        pass
    else:
        return signature, return_annotation


def setup(app):
    app.connect('autodoc-process-signature', autodoc_process_signature)



# coding: utf-8
from __future__ import absolute_import, division, print_function

from collections import Iterable

import six


class UnsetType(object):
    def __repr__(self):
        return 'UNSET'


UNSET = UnsetType()
del UnsetType


class Infinity(object):

    def __repr__(self):
        return "Infinity"

    def __hash__(self):
        return hash(repr(self))

    def __lt__(self, other):
        return False

    def __le__(self, other):
        return False

    def __eq__(self, other):
        return isinstance(other, self.__class__)

    def __ne__(self, other):
        return not isinstance(other, self.__class__)

    def __gt__(self, other):
        return True

    def __ge__(self, other):
        return True

    def __neg__(self):
        return NegativeInfinity


Infinity = Infinity()


class NegativeInfinity(object):

    def __repr__(self):
        return "-Infinity"

    def __hash__(self):
        return hash(repr(self))

    def __lt__(self, other):
        return True

    def __le__(self, other):
        return True

    def __eq__(self, other):
        return isinstance(other, self.__class__)

    def __ne__(self, other):
        return not isinstance(other, self.__class__)

    def __gt__(self, other):
        return False

    def __ge__(self, other):
        return False

    def __neg__(self):
        return Infinity


NegativeInfinity = NegativeInfinity()


def fixup_module_metadata(module_name, namespace):
    def fix_one(obj):
        mod = getattr(obj, "__module__", None)
        if mod is not None and mod.startswith("outcome."):
            obj.__module__ = module_name
            if isinstance(obj, type):
                for attr_value in obj.__dict__.values():
                    fix_one(attr_value)

    for objname in namespace["__all__"]:
        obj = namespace[objname]
        fix_one(obj)


def force_tuple(n):
    if isinstance(n, six.string_types):
        raise TypeError('Expected tuple or int.')
    if not isinstance(n, Iterable):
        return n,
    if not isinstance(n, tuple):
        return tuple(n)
    return n


def doc_signature(signature):
    def decorator(fn):
        fn.__parverdoc_signature__ = signature
        return fn
    return decorator


def kwonly_args(kws, required, withdefaults=(), leftovers=False):
    """

    Based on the snippet by Eric Snow
    http://code.activestate.com/recipes/577940

    SPDX-License-Identifier: MIT
    """

    if hasattr(withdefaults, "items"):
        # allows for OrderedDict to be passed
        withdefaults = withdefaults.items()

    kwonly = []

    # extract the required keyword-only arguments
    missing = []
    for name in required:
        if name not in kws:
            missing.append(name)
        else:
            kwonly.append(kws.pop(name))

    # validate required keyword-only arguments
    if missing:
        if len(missing) > 2:
            end = "s: %s, and %s" % (", ".join(missing[:-1]), missing[-1])
        elif len(missing) == 2:
            end = "s: %s and %s" % tuple(missing)
        else:
            end = ": %s" % tuple(missing)

        msg = "missing %s required keyword-only argument%s"
        raise TypeError(msg % (len(missing), end))

    # handle the withdefaults
    for name, value in withdefaults:
        if name not in kws:
            kwonly.append(value)
        else:
            kwonly.append(kws.pop(name))

    # handle any leftovers
    if not leftovers and kws:
        msg = "got an unexpected keyword argument '%s'"
        raise TypeError(msg % (kws.keys()[0]))

    return [kws] + kwonly



# coding: utf-8
# This file is imported from __init__.py and exec'd from setup.py

__version__ = '0.2.1'
__description__ = 'Parse and manipulate version numbers.'

__license__ = 'MIT'

__author__ = 'Frazer McLean'
__email__ = 'frazer@frazermclean.co.uk'



# coding: utf-8
from __future__ import absolute_import, division, print_function

from ._about import __author__, __email__, __version__  # noqa
from ._parse import ParseError
from ._version import Version

__all__ = ('Version', 'ParseError')

from ._helpers import fixup_module_metadata
fixup_module_metadata(__name__, globals())
del fixup_module_metadata



# coding: utf-8
from __future__ import absolute_import, division, print_function

import attr
import six
from arpeggio import NoMatch, PTNodeVisitor, visit_parse_tree, Terminal
from arpeggio.cleanpeg import ParserPEG

from ._helpers import UNSET
from . import _segments as segment

canonical = r'''
    version = epoch? release pre? post? dev? local? EOF
    epoch = int "!"
    release = int (dot int)*
    pre = pre_tag pre_post_num
    pre_tag = "a" / "b" / "rc"
    post = sep post_tag pre_post_num
    pre_post_num = int
    post_tag = "post"
    dev = sep "dev" int
    local = "+" r'([a-zA-Z0-9]+([-_\.][a-zA-Z0-9]+)*)'
    sep = dot
    dot = "."
    int = r'[0-9]+'
    alpha = r'[a-zA-Z0-9]'
'''

permissive = r'''
    version = v? epoch? release pre? (post / post_implicit)? dev? local? EOF
    v = "v"
    epoch = int "!"
    release = int (dot int)*
    pre = sep? pre_tag pre_post_num?
    pre_tag = "c" / "rc" / "alpha" / "a" / "beta" / "b" / "preview" / "pre"
    post = sep? post_tag pre_post_num?
    post_implicit = "-" int
    post_tag = "post" / "rev" / "r"
    pre_post_num = sep? int
    dev = sep? "dev" int?
    local = "+" r'([a-zA-Z0-9]+([-_\.][a-zA-Z0-9]+)*)'
    sep = dot / "-" / "_"
    dot = "."
    int = r'[0-9]+'
    alpha = r'[a-zA-Z0-9]'
'''

_strict_parser = ParserPEG(canonical, root_rule_name='version', skipws=False)
_permissive_parser = ParserPEG(
    permissive, root_rule_name='version', skipws=False, ignore_case=True)


def unwrap_token(value):
    if isinstance(value, Token):
        return value.value
    return value


@attr.s
class Token(object):
    value = attr.ib()


def make_token(name):
    return type(name, (Token,), dict())


Sep = make_token('Sep')
Tag = make_token('Tag')
VToken = make_token('VToken')


class VersionVisitor(PTNodeVisitor):
    def visit_version(self, node, children):
        return children

    def visit_v(self, node, children):
        return segment.V()

    def visit_epoch(self, node, children):
        return segment.Epoch(children[0])

    def visit_release(self, node, children):
        return segment.Release(tuple(children))

    def visit_pre(self, node, children):
        sep1 = UNSET
        tag = UNSET
        sep2 = UNSET
        num = UNSET

        for token in children:
            if sep1 is UNSET:
                if isinstance(token, Sep):
                    sep1 = token.value
                elif isinstance(token, Tag):
                    sep1 = None
                    tag = token.value
            elif tag is UNSET:
                tag = token.value
            else:
                assert isinstance(token, tuple)
                assert len(token) == 2
                sep2 = token[0].value
                num = token[1]

        if sep2 is UNSET:
            sep2 = None
            num = None

        assert sep1 is not UNSET
        assert tag is not UNSET
        assert sep2 is not UNSET
        assert num is not UNSET

        return segment.Pre(sep1=sep1, tag=tag, sep2=sep2, value=num)

    def visit_pre_post_num(self, node, children):
        # when "pre_post_num = int", visit_int isn't called for some reason
        # I don't understand. Let's call int() manually
        if isinstance(node, Terminal):
            return Sep(None), int(node.value)

        if len(children) == 1:
            return Sep(None), children[0]
        else:
            return tuple(children[:2])

    def visit_pre_tag(self, node, children):
        return Tag(node.value)

    def visit_post(self, node, children):
        sep1 = UNSET
        tag = UNSET
        sep2 = UNSET
        num = UNSET

        for token in children:
            if sep1 is UNSET:
                if isinstance(token, Sep):
                    sep1 = token.value
                elif isinstance(token, Tag):
                    sep1 = None
                    tag = token.value
            elif tag is UNSET:
                tag = token.value
            else:
                assert isinstance(token, tuple)
                assert len(token) == 2
                sep2 = token[0].value
                num = token[1]

        if sep2 is UNSET:
            sep2 = None
            num = None

        assert sep1 is not UNSET
        assert tag is not UNSET
        assert sep2 is not UNSET
        assert num is not UNSET

        return segment.Post(sep1=sep1, tag=tag, sep2=sep2, value=num)

    def visit_post_tag(self, node, children):
        return Tag(node.value)

    def visit_post_implicit(self, node, children):
        return segment.Post(sep1=UNSET, tag=None, sep2=UNSET, value=children[0])

    def visit_dev(self, node, children):
        num = None
        sep = UNSET

        for token in children:
            if sep is UNSET:
                if isinstance(token, Sep):
                    sep = token.value
                else:
                    num = token
            else:
                num = token

        if sep is UNSET:
            sep = None

        return segment.Dev(value=num, sep=sep)

    def visit_local(self, node, children):
        return segment.Local(''.join(children))

    def visit_int(self, node, children):
        return int(node.value)

    def visit_sep(self, node, children):
        return Sep(node.value)


class ParseError(ValueError):
    """Raised when parsing an invalid version number."""


def parse(version, strict=False):
    parser = _strict_parser if strict else _permissive_parser

    if not strict:
        version = version.lower()

    try:
        tree = parser.parse(version.strip())
    except NoMatch as exc:
        six.raise_from(ParseError(str(exc)), None)

    return visit_parse_tree(tree, VersionVisitor())



# coding: utf-8
from __future__ import absolute_import, division, print_function

import pytest
from hypothesis import assume, given

from parver import ParseError, Version

from .strategies import version_string, whitespace


@given(whitespace, version_string(), whitespace)
def test_parse_hypothesis(prefix, version, suffix):
    Version.parse(prefix + version + suffix)


@given(whitespace, version_string(strict=True), whitespace)
def test_parse_strict_hypothesis(prefix, version, suffix):
    Version.parse(prefix + version + suffix, strict=True)


@given(version_string(strict=False))
def test_parse_strict_error(version):
    v = Version.parse(version)

    # Exclude already normalized versions
    assume(str(v.normalize()) != version)

    # 0!1 normalizes to '1'
    assume(v.epoch != 0 or v.epoch_implicit)

    with pytest.raises(ParseError):
        Version.parse(version, strict=True)


@given(version_string())
def test_roundtrip(version):
    assert str(Version.parse(version)) == version



# coding: utf-8
from __future__ import absolute_import, division, print_function

import string

from parver import Version

from hypothesis.strategies import (
    composite, integers, just, lists, one_of, sampled_from, text)

num_int = integers(min_value=0)
num_str = num_int.map(str)


def epoch():
    epoch = num_str.map(lambda s: s + '!')
    return one_of(just(''), epoch)


@composite
def release(draw):
    a = draw(num_str)
    parts = [a] + draw(lists(num_str.map(lambda s: '.' + s)))
    return ''.join(parts)


def separator(strict=False, optional=False):
    sep = ['.']

    if optional:
        sep.append('')

    if not strict:
        sep.extend(['-', '_'])

    return sampled_from(sep)


@composite
def pre(draw, strict=False):
    words = ['a', 'b', 'rc']
    if not strict:
        words.extend(['c', 'alpha', 'beta', 'pre', 'preview'])

    sep1 = draw(separator(strict=strict, optional=True))
    if strict:
        sep1 = ''

    word = draw(sampled_from(words))

    if strict:
        sep2 = ''
    else:
        sep2 = draw(separator(strict=strict, optional=True))

    n = draw(num_str)

    if strict:
        num_part = draw(just(sep2 + n))
    else:
        num_part = draw(one_of(just(''), just(sep2 + n)))

    return draw(one_of(just(''), just(sep1 + word + num_part)))


@composite
def post(draw, strict=False):
    words = ['post']
    if not strict:
        words.extend(['r', 'rev'])

    sep1 = draw(separator(strict=strict, optional=not strict))
    word = draw(sampled_from(words))

    sep2 = draw(separator(strict=strict, optional=True))
    if strict:
        sep2 = ''

    n = draw(num_str)

    if strict:
        num_part = draw(just(sep2 + n))
    else:
        num_part = draw(one_of(just(''), just(sep2 + n)))

    post = sep1 + word + num_part

    if strict:
        return post

    post = just(post)
    post_implicit = num_str.map(lambda s: '-' + s)

    return draw(one_of(just(''), post_implicit, post))


@composite
def dev(draw, strict=False):
    sep = draw(separator(strict=strict, optional=not strict))

    if strict:
        num_part = draw(num_str)
    else:
        num_part = draw(one_of(just(''), num_str))

    return draw(one_of(just(''), just(sep + 'dev' + num_part)))


@composite
def local_segment(draw, strict=False):
    if strict:
        sep = just('.')
    else:
        sep = sampled_from('-_.')

    return draw(sep) + draw(text(string.ascii_lowercase + string.digits, min_size=1))


@composite
def local(draw, strict=False):
    start = draw(text(string.ascii_lowercase + string.digits, min_size=1))
    end = ''.join(draw(lists(local_segment(strict=strict))))

    return draw(one_of(just(''), just('+' + start + end)))


whitespace = sampled_from(['', '\t', '\n', '\r', '\f', '\v'])


def vchar(strict=False):
    if strict:
        return just('')
    return sampled_from(['', 'v'])


@composite
def version_string(draw, strict=False):
    return (
        draw(vchar(strict=strict)) +
        draw(epoch()) +
        draw(release()) +
        draw(pre(strict=strict)) +
        draw(post(strict=strict)) +
        draw(dev(strict=strict)) +
        draw(local(strict=strict))
    )


@composite
def version_strategy(draw, strict=False):
    return Version.parse(draw(version_string(strict=strict)))



# coding: utf-8
from __future__ import absolute_import, division, print_function

from io import open

from setuptools import setup, find_packages

about = dict()

# read _version.py as bytes, otherwise exec will complain about
# 'coding: utf-8', which we want there for the normal Python 2 import
with open('src/parver/_about.py', 'rb') as fp:
    about_mod = fp.read()

exec(about_mod, about)

LONG_DESC = open('README.rst', encoding='utf-8').read()

extras_require = {
    'test': [
        'pytest ~= 3.5',
        'hypothesis ~= 3.56',
        'pretend ~= 1.0',
    ],
    'docstest': [
        'doc8',
        'sphinx',
        'sphinx_rtd_theme',
    ],
    'pep8test': [
        'flake8',
        'pep8-naming',
    ],
}

setup(
    name='parver',
    version=about['__version__'],
    description=about['__description__'],
    url='https://github.com/RazerM/parver',
    project_urls={
        "Documentation": "https://parver.readthedocs.io",
    },
    long_description=LONG_DESC,
    long_description_content_type='text/x-rst',
    author=about['__author__'],
    author_email=about['__email__'],
    license=about['__license__'],
    packages=find_packages('src'),
    package_dir={'': 'src'},
    install_requires=[
        'arpeggio ~= 1.7',
        'attrs >= 17.4.0',
        'six ~= 1.9',
    ],
    extras_require=extras_require,
    python_requires='>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*',
    keywords='pep440 version parse',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
    ],
)



# coding: utf-8
from __future__ import absolute_import, division, print_function

import pytest
from hypothesis import given

from parver import Version

from .strategies import version_strategy


def v(*args, **kwargs):
    return args, kwargs


@pytest.mark.parametrize('vargs, s', [
    (v(1), '1'),
    (v(release=(1,)), '1'),
    (v(release=(1, 2)), '1.2'),
    # epoch
    (v(release=1, epoch=None), '1'),
    (v(release=1, epoch=0), '0!1'),
    (v(release=1, epoch=1), '1!1'),
    (v(release=1, pre_tag=None), '1'),
    # pre_tag without pre
    (v(release=1, pre_tag='c'), '1c'),
    (v(release=1, pre_tag='rc'), '1rc'),
    (v(release=1, pre_tag='alpha'), '1alpha'),
    (v(release=1, pre_tag='a'), '1a'),
    (v(release=1, pre_tag='beta'), '1beta'),
    (v(release=1, pre_tag='b'), '1b'),
    (v(release=1, pre_tag='preview'), '1preview'),
    (v(release=1, pre_tag='pre'), '1pre'),
    # pre_tag with pre
    (v(release=1, pre=0, pre_tag='c'), '1c0'),
    (v(release=1, pre=1, pre_tag='rc'), '1rc1'),
    (v(release=1, pre=2, pre_tag='alpha'), '1alpha2'),
    (v(release=1, pre=3, pre_tag='a'), '1a3'),
    (v(release=1, pre=None, pre_tag='beta'), '1beta'),
    (v(release=1, pre=0, pre_tag='b'), '1b0'),
    (v(release=1, pre=0, pre_tag='preview'), '1preview0'),
    (v(release=1, pre=0, pre_tag='pre'), '1pre0'),
    # pre_tag with pre_sep1
    (v(release=1, pre_sep1=None, pre_tag='b'), '1b'),
    (v(release=1, pre_sep1='.', pre_tag='b'), '1.b'),
    (v(release=1, pre_sep1='-', pre_tag='b'), '1-b'),
    (v(release=1, pre_sep1='_', pre_tag='b'), '1_b'),
    # pre_tag with pre_sep2
    (v(release=2, pre=1, pre_sep2=None, pre_tag='b'), '2b1'),
    (v(release=2, pre=1, pre_sep2='.', pre_tag='b'), '2b.1'),
    (v(release=2, pre=1, pre_sep2='-', pre_tag='b'), '2b-1'),
    (v(release=2, pre=1, pre_sep2='_', pre_tag='b'), '2b_1'),
    # pre_tag with pre_sep1 and pre_sep2
    (v(release=2, pre=1, pre_sep1='.', pre_sep2=None, pre_tag='b'), '2.b1'),
    (v(release=2, pre=1, pre_sep1='.', pre_sep2='.', pre_tag='b'), '2.b.1'),
    (v(release=2, pre=1, pre_sep1='.', pre_sep2='-', pre_tag='b'), '2.b-1'),
    (v(release=2, pre=1, pre_sep1='.', pre_sep2='_', pre_tag='b'), '2.b_1'),
    # post
    (v(release=1, post=None), '1.post'),
    (v(release=1, post=0), '1.post0'),
    (v(release=1, post=1), '1.post1'),
    # post_tag
    (v(release=1, post=0, post_tag=None), '1-0'),
    (v(release=1, post=None, post_tag='post'), '1.post'),
    (v(release=1, post=None, post_tag='r'), '1.r'),
    (v(release=1, post=None, post_tag='rev'), '1.rev'),
    (v(release=1, post_tag='post'), '1.post'),
    (v(release=1, post_tag='r'), '1.r'),
    (v(release=1, post_tag='rev'), '1.rev'),
    # post with post_sep1
    (v(release=1, post=None, post_sep1=None), '1post'),
    (v(release=1, post=None, post_sep1='.'), '1.post'),
    (v(release=1, post=None, post_sep1='-'), '1-post'),
    (v(release=1, post=None, post_sep1='_'), '1_post'),
    # post with post_sep2
    (v(release=2, post=1, post_sep2=None), '2.post1'),
    (v(release=2, post=1, post_sep2='.'), '2.post.1'),
    (v(release=2, post=1, post_sep2='-'), '2.post-1'),
    (v(release=2, post=1, post_sep2='_'), '2.post_1'),
    # post with post_sep1 and post_sep2
    (v(release=2, post=1, post_sep1='.', post_sep2=None), '2.post1'),
    (v(release=2, post=1, post_sep1='.', post_sep2='.'), '2.post.1'),
    (v(release=2, post=1, post_sep1='.', post_sep2='-'), '2.post-1'),
    (v(release=2, post=1, post_sep1='.', post_sep2='_'), '2.post_1'),
    # dev
    (v(release=1, dev=None), '1.dev'),
    (v(release=1, dev=0), '1.dev0'),
    (v(release=1, dev=1), '1.dev1'),
    # local
    (v(release=1, local=None), '1'),
    (v(release=1, local='a'), '1+a'),
    (v(release=1, local='0'), '1+0'),
    (v(release=1, local='a0'), '1+a0'),
    (v(release=1, local='a.0'), '1+a.0'),
    (v(release=1, local='0-0'), '1+0-0'),
    (v(release=1, local='0_a'), '1+0_a'),
])
def test_init(vargs, s):
    args, kwargs = vargs
    assert str(Version(*args, **kwargs)) == s


@pytest.mark.parametrize('kwargs', [
    dict(pre=1),
    dict(pre_sep1='.'),
    dict(pre_sep2='.'),
    dict(pre_sep1='.', pre_sep2='.'),
    dict(post_tag=None),
    dict(post_tag=None, post=None),
    dict(post_tag=None, post_sep1='.'),
    dict(post_tag=None, post_sep2='.'),
    dict(post_tag=None, post_sep1='.', post_sep2='.'),
])
def test_invalid(kwargs):
    """Test bad keyword combinations."""
    with pytest.raises(ValueError):
        Version(release=1, **kwargs)


@pytest.mark.parametrize('kwargs', [
    dict(release='1'),
    dict(v=3),
    dict(epoch='1'),
    dict(pre_tag='b', pre='2'),
    dict(post='3'),
    dict(dev='3'),
    dict(local=[1, 'abc']),
    dict(local=1),
])
def test_validation_type(kwargs):
    if 'release' not in kwargs:
        kwargs['release'] = 1

    with pytest.raises(TypeError):
        # print so we can see output when test fails
        print(Version(**kwargs))


@pytest.mark.parametrize('kwargs', [
    dict(pre_tag='alph'),
    dict(pre_tag='a', pre_sep1='x'),
    dict(pre_tag='a', pre_sep2='x'),
    dict(post=1, post_sep1='x'),
    dict(post=1, post_sep2='x'),
    dict(dev=4, dev_sep='y'),
    dict(post_tag=None, post=1, post_sep1='.'),
    dict(post_tag=None, post=1, post_sep2='.'),
])
def test_validation_value(kwargs):
    if 'release' not in kwargs:
        kwargs['release'] = 1

    with pytest.raises(ValueError):
        # print so we can see output when test fails
        print(Version(**kwargs))


@pytest.mark.parametrize('kwargs, values, version', [
    (
        dict(release=1),
        dict(
            release=(1,),
            v=False,
            epoch=0,
            epoch_implicit=True,
            pre_tag=None,
            pre=None,
            pre_implicit=False,
            pre_sep1=None,
            pre_sep2=None,
            post=None,
            post_tag=None,
            post_implicit=False,
            post_sep1=None,
            post_sep2=None,
            dev=None,
            dev_implicit=False,
            dev_sep=None,
            local=None,
        ),
        '1',
    ),
    (
        dict(epoch=0),
        dict(epoch=0, epoch_implicit=False),
        '0!1',
    ),
    (
        dict(pre=None, pre_tag='b'),
        dict(
            pre=0,
            pre_tag='b',
            pre_implicit=True,
            pre_sep1=None,
            pre_sep2=None,
        ),
        '1b',
    ),
    (
        dict(pre=0, pre_tag='a'),
        dict(
            pre=0,
            pre_tag='a',
            pre_implicit=False,
            pre_sep1=None,
            pre_sep2=None,
        ),
        '1a0',
    ),
    (
        dict(pre=2, pre_tag='pre', pre_sep1='-', pre_sep2='.'),
        dict(
            pre=2,
            pre_tag='pre',
            pre_implicit=False,
            pre_sep1='-',
            pre_sep2='.',
        ),
        '1-pre.2',
    ),
    (
        dict(post=None),
        dict(
            post=0,
            post_tag='post',
            post_implicit=True,
            post_sep1='.',
            post_sep2=None,
        ),
        '1.post',
    ),
    (
        dict(post=None, post_sep2='.'),
        dict(
            post=0,
            post_tag='post',
            post_implicit=True,
            post_sep1='.',
            post_sep2='.',
        ),
        '1.post.',
    ),
    (
        dict(post=0),
        dict(
            post=0,
            post_tag='post',
            post_implicit=False,
            post_sep1='.',
            post_sep2=None,
        ),
        '1.post0',
    ),
    (
        dict(post=0, post_tag=None),
        dict(
            post=0,
            post_tag=None,
            post_implicit=False,
            post_sep1='-',
            post_sep2=None,
        ),
        '1-0',
    ),
    (
        dict(post=3, post_tag='rev', post_sep1='-', post_sep2='_'),
        dict(
            post=3,
            post_tag='rev',
            post_implicit=False,
            post_sep1='-',
            post_sep2='_',
        ),
        '1-rev_3',
    ),
    (
        dict(dev=None),
        dict(dev=0, dev_implicit=True, dev_sep='.'),
        '1.dev',
    ),
    (
        dict(dev=2),
        dict(dev=2, dev_implicit=False, dev_sep='.'),
        '1.dev2',
    ),
    (
        dict(dev=0, dev_sep='-'),
        dict(dev=0, dev_implicit=False, dev_sep='-'),
        '1-dev0',
    ),
    (
        dict(local='a.b'),
        dict(local='a.b'),
        '1+a.b',
    ),
])
def test_attributes(kwargs, values, version):
    # save us repeating ourselves in test data above
    kwargs.setdefault('release', 1)

    v = Version(**kwargs)
    assert str(v) == version
    for key, value in values.items():
        assert getattr(v, key) == value


@given(version_strategy())
def test_replace_roundtrip(version):
    """All the logic inside replace() is in converting the attributes to the
    form expected by __init__, so this function tests most of that.
    """
    assert version.replace() == version


@pytest.mark.parametrize('before, kwargs, after', [
    (
        'v0!1.2.alpha-3_rev.4_dev5+l.6',
        dict(
            release=(2, 1),
            epoch=None,
            v=False,
            pre_tag='a',
            pre_sep1=None,
            pre_sep2=None,
            post_tag='post',
            post_sep1='.',
            post_sep2=None,
            dev_sep='.',
            local=None
        ),
        '2.1a3.post4.dev5',
    ),
    (
        '2.1a3.post4.dev5',
        dict(
            release=(1, 2),
            epoch=0,
            v=True,
            pre_tag='alpha',
            pre_sep1='.',
            pre_sep2='-',
            post_tag='rev',
            post_sep1='_',
            post_sep2='.',
            dev_sep='_',
            local='l.6'
        ),
        'v0!1.2.alpha-3_rev.4_dev5+l.6',
    ),
    (
        '2.post4',
        dict(post_tag=None),
        '2-4',
    ),
])
def test_replace(before, kwargs, after):
    """Make sure the keys we expect are passed through."""
    assert str(Version.parse(before).replace(**kwargs)) == after


@pytest.mark.parametrize('before, kwargs, after', [
    (
        '1.2.alpha-3_rev.4_dev5',
        dict(pre=True, post=True, dev=True),
        '1.2',
    ),
    (
        '1.2.alpha-3_rev.4_dev5',
        dict(pre=False, post=False, dev=False),
        '1.2.alpha-3_rev.4_dev5',
    ),
    (
        '1.2.alpha-3_rev.4_dev5',
        dict(),
        '1.2.alpha-3_rev.4_dev5',
    ),
    (
        '1.2',
        dict(pre=True, post=True, dev=True),
        '1.2',
    ),
])
def test_clear(before, kwargs, after):
    assert str(Version.parse(before).clear(**kwargs)) == after


@pytest.mark.parametrize('before, index, after', [
    ('1', 0, '2'),
    ('1', 1, '1.1'),
    ('1', 2, '1.0.1'),
    ('1.1', 0, '2.0'),
    ('1.1', 1, '1.2'),
    ('1.1', 2, '1.1.1'),
    ('1.1', 3, '1.1.0.1'),
    ('4.3.2.1', 2, '4.3.3.0'),
])
def test_bump_release(before, index, after):
    assert str(Version.parse(before).bump_release(index=index)) == after


@pytest.mark.parametrize('before, index, value, after', [
    ('2', 0, 1, '1'),
    ('2', 0, 2, '2'),
    ('2', 0, 3, '3'),
    ('2', 1, 3, '2.3'),
    ('2', 2, 3, '2.0.3'),
    ('2.4', 0, 4, '4.0'),
    ('2.4', 1, 6, '2.6'),
    ('2.4', 2, 6, '2.4.6'),
    ('2.4', 3, 6, '2.4.0.6'),
    ('4.3.2.1', 1, 5, '4.5.0.0'),
    # e.g. CalVer
    ('2017.4', 0, 2018, '2018.0'),
    ('17.5.1', 0, 18, '18.0.0'),
    ('18.0.0', 1, 2, '18.2.0'),
])
def test_bump_release_to(before, index, value, after):
    v = Version.parse(before).bump_release_to(index=index, value=value)
    assert str(v) == after


@pytest.mark.parametrize('before, index, value, after', [
    ('2', 0, 1, '1'),
    ('2', 0, 2, '2'),
    ('2', 0, 3, '3'),
    ('2', 1, 3, '2.3'),
    ('2', 2, 3, '2.0.3'),
    ('2.4', 0, 4, '4.4'),
    ('2.4', 1, 6, '2.6'),
    ('2.4', 2, 6, '2.4.6'),
    ('2.4', 3, 6, '2.4.0.6'),
    ('2.0.4', 1, 3, '2.3.4'),
    ('4.3.2.1', 1, 5, '4.5.2.1'),
])
def test_set_release(before, index, value, after):
    v = Version.parse(before).set_release(index=index, value=value)
    assert str(v) == after


@pytest.mark.parametrize('index, exc', [
    ('1', TypeError),
    (1.1, TypeError),
    (-1, ValueError),
])
def test_bump_release_error(index, exc):
    with pytest.raises(exc):
        print(Version(release=1).bump_release(index=index))


@pytest.mark.parametrize('before, tag, after', [
    ('1', 'a', '1a0'),
    ('1a0', None, '1a1'),
    ('1a', None, '1a1'),
    ('1a', 'a', '1a1'),
    ('1.b-0', None, '1.b-1'),
])
def test_bump_pre(before, tag, after):
    assert str(Version.parse(before).bump_pre(tag)) == after


@pytest.mark.parametrize('version, tag', [
    ('1.2', None),
    ('1.2a', 'b'),
])
def test_bump_pre_error(version, tag):
    with pytest.raises(ValueError):
        print(Version.parse(version).bump_pre(tag))


@pytest.mark.parametrize('before, kwargs, after', [
    ('1', dict(), '1.post0'),
    ('1.post0', dict(), '1.post1'),
    ('1rev', dict(), '1rev1'),
    ('1-0', dict(), '1-1'),
    ('1-0', dict(tag='post'), '1.post1'),
    ('1-post_0', dict(tag=None), '1-1'),
])
def test_bump_post(before, kwargs, after):
    assert str(Version.parse(before).bump_post(**kwargs)) == after


@pytest.mark.parametrize('before, after', [
    ('1', '1.dev0'),
    ('1.dev0', '1.dev1'),
    ('1-dev1', '1-dev2'),
])
def test_bump_dev(before, after):
    assert str(Version.parse(before).bump_dev()) == after


@pytest.mark.parametrize('arg, expected', [
    (1, (1,)),
    ([1], (1,)),
    ((1, 2), (1, 2)),
    ([1, 2], (1, 2)),
    # range is a Sequence
    (range(1, 3), (1, 2)),
    # An iterable that is not also a sequence
    ((x for x in range(1, 3)), (1, 2)),
])
def test_release_tuple(arg, expected):
    v = Version(release=arg)
    assert isinstance(v.release, tuple)
    assert v.release == expected


@pytest.mark.parametrize('version', [
    '1a',
    '1alpha',
    '1a1',
])
def test_is_alpha(version):
    v = Version.parse(version)
    assert v.is_alpha
    assert not v.is_beta
    assert not v.is_release_candidate


@pytest.mark.parametrize('version', [
    '1b',
    '1beta',
    '1b1',
])
def test_is_beta(version):
    v = Version.parse(version)
    assert not v.is_alpha
    assert v.is_beta
    assert not v.is_release_candidate


@pytest.mark.parametrize('version', [
    '1rc',
    '1c',
    '1pre',
    '1preview',
    '1rc1',
])
def test_is_release_candidate(version):
    v = Version.parse(version)
    assert not v.is_alpha
    assert not v.is_beta
    assert v.is_release_candidate


def test_ambiguous():
    with pytest.raises(ValueError, match='post_tag.*pre'):
        Version(release=1, pre_tag='rc', post=2, post_tag=None)



# coding: utf-8
from __future__ import absolute_import, division, print_function

import attr


@attr.s(slots=True)
class Segment(object):
    pass


@attr.s(slots=True)
class V(Segment):
    pass


@attr.s(slots=True)
class ValueSegment(Segment):
    value = attr.ib()


@attr.s(slots=True)
class Epoch(ValueSegment):
    pass


@attr.s(slots=True)
class Release(ValueSegment):
    pass


@attr.s(slots=True)
class Pre(ValueSegment):
    sep1 = attr.ib()
    tag = attr.ib()
    sep2 = attr.ib()


@attr.s(slots=True)
class Post(ValueSegment):
    sep1 = attr.ib()
    tag = attr.ib()
    sep2 = attr.ib()


@attr.s(slots=True)
class Dev(ValueSegment):
    sep = attr.ib()


@attr.s(slots=True)
class Local(ValueSegment):
    pass
